declare namespace Httpr {

    /**
     * String hash map.
     */

    export type StringMap = {
        [key: string]: string;
    };

    /**
     * Available HTTP methods.
     *
     * HTTP methods are:
     * - get
     * - post
     * - put
     * - delete
     * - options
     * - patch
     * - head
     * - trace
     * - connect
     */

    export type HttpMethod = 'get' | 'post' | 'put' | 'delete' | 'options' | 'patch' | 'head' | 'trace' | 'connect';

    export type Primitives = string | number | boolean;

    /**
     * One-level depth objects with simple values.
     */

    export type PlainObject = {
        [key: string]: Primitives | Array<Primitives>;
    };

    /**
     * Http Method constants.
     *
     * @see {@link HttpMethod}
     */

    export class HttpMethods {
        static GET: HttpMethod;
        static POST: HttpMethod;
        static PUT: HttpMethod;
        static DELETE: HttpMethod;
        static OPTIONS: HttpMethod;
        static PATCH: HttpMethod;
        static HEAD: HttpMethod;
        static TRACE: HttpMethod;
        static CONNECT: HttpMethod;
    }

    /**
     * Http Request Settings.
     */

    export type HttpRequestSettings = {

        /**
         * Request method.
         */
        method: HttpMethod;

        /**
         * Resource url.
         */
        url: string;

        /**
         * Query parameters.
         */
        params: PlainObject;

        /**
         * Request headers.
         */
        headers: StringMap;

        /**
         * Request body.
         */
        body?: any;
    };

    /**
     * Http Response.
     */

    export type HttpResponse = {

        /**
         * HTTP status code.
         */
        status: number;

        /**
         * HTTP status code text.
         */
        statusText: string;

        /**
         * Raw response body.
         */
        responseText: string;

        /**
         * Processed response data.
         */
        data?: any;

        /**
         * Response headers.
         */
        headers: StringMap;
    };

    /**
     * Httpr Interceptor default implementation.
     *
     * This class should be extended to create new, functional interceptors.
     */

    export class HttprInterceptor {

        /**
         * Apply modifications to request settings.
         *
         * @param {HttpRequestSettings} settings - current request settings.
         * @return {HttpRequestSettings} modified request settings.
         */
        beforeRequest(settings: HttpRequestSettings): HttpRequestSettings;

        /**
         * Apply modifications to success response.
         *
         * @param {HttpResponse} response - success response.
         * @return {HttpResponse} modified response.
         */
        afterSuccess(response: HttpResponse): HttpResponse;

        /**
         * Apply modifications to error response.
         *
         * @param {HttpResponse} response - error response.
         * @return {HttpResponse} modified response.
         */
        afterError(response: HttpResponse): HttpResponse;
    }

    /**
     * Httpr Provider fake implementation.
     *
     * This class should be extended and overridden by a low level implementation to perform
     * the actual request.
     */

    export class HttprProvider {

        /**
         * Perform a request with low level implementation.
         *
         * @param {HttpRequestSettings} settings - request settings.
         * @return {Promise<HttpResponse>} promise handler.
         */
        request(settings: HttpRequestSettings): Promise<HttpResponse>;
    }

    /**
     * Httpr Configurations.
     */

    export type HttprConfig = {

        /**
         * A base URL to append to all requests.
         */
        baseUrl?: string;

        /**
         * Implementation of the provider that will perform the requests.
         */
        provider: HttprProvider;
    };

    /**
     * Http Header constants.
     *
     * Bundled headers:
     * - accept
     * - content-type
     */

    export class HttpHeaders {
        static ACCEPT: string;
        static ACCEPT_CHARSET: string;
        static ACCEPT_ENCODING: string;
        static ACCEPT_LANGUAGE: string;
        static ACCEPT_DATETIME: string;
        static ACCEPT_RANGES: string;
        static ACCESS_CONTROL_ALLOW_ORIGIN: string;
        static AUTHORIZATION: string;
        static CACHE_CONTROL: string;
        static CONNECTION: string;
        static COOKIE: string;
        static CONTENT_ENCODING: string;
        static CONTENT_LENGTH: string;
        static CONTENT_MD5: string;
        static CONTENT_TYPE: string;
        static DATE: string;
        static ETAG: string;
        static EXPECT: string;
        static EXPIRES: string;
        static FORWARDED: string;
        static FROM: string;
        static HOST: string;
        static PERMANENT: string;
        static IF_MATCH: string;
        static IF_MODIFIED_SINCE: string;
        static IF_NONE_MATCH: string;
        static IF_RANGE: string;
        static IF_UNMODIFIED_SINCE: string;
        static LAST_MODIFIED: string;
        static MAX_FORWARDS: string;
        static ORIGIN: string;
        static PRAGMA: string;
        static PROXY_AUTHORIZATION: string;
        static RANGE: string;
        static REFERER: string;
        static STATUS: string;
        static TE: string;
        static USER_AGENT: string;
        static UPGRADE: string;
        static VIA: string;
        static WARNING: string;
    }

    /**
     * Media Type constants.
     *
     * Bundled Media Types:
     * - application/json
     * - application/x-www-form-urlencoded
     * - text/plain
     */

    export class MediaTypes {
        static APPLICATION_JSON: string;
        static FORM_URLENCODED: string;
        static TEXT_PLAIN: string;
    }

    /**
     * Encode plain object into query string.
     *
     * @param {PlainObject} object - hash of items.
     * @returns {string} encoded query string.
     */
    export function urlEncode(object: PlainObject): string;

    /**
     * Decode a query string to plain object.
     *
     * @param {string} queryParams - encoded query string.
     * @return {PlainObject} hash of items.
     */
    export function urlDecode(queryParams: string): PlainObject;

    /**
     * Join two URL parts.
     *
     * @param {Array<string>} parts - URL parts.
     * @return {string} joined URL.
     */
    export function urlJoin(...parts: Array<string>): string;

    /**
     * Custom body especification for Httpr.
     */
    export abstract class HttprBody<T> {
        protected _type: string;
        protected _content: T;

        /**
         * Create a custom body type.
         *
         * @param {string} type - media type value.
         * @param {T} content - raw body content.
         */
        constructor(type: string, content: T);

        /**
         * Retrieve body media type.
         *
         * @return {string} content type header value.
         */
        readonly type: string;

        /**
         * Retrieve raw body content.
         *
         * @return {T} body content.
         */
        readonly content: T;

        /**
         * Serialize function for body content.
         *
         * @returns {string} serialized body.
         */
        abstract toString(): string;
    }

    /**
     * Static methods used in Httpr for preparing requests and managing interceptors.
     */

    export class HttprStatic {

        /**
         * Prepare request settings and apply interceptors before sending the request.
         *
         * @param {Httpr} instance - instance of Http
         * @param {HttpMethod} method - http method to use.
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @param {*} body - request body data.
         * @returns {HttpRequestSettings} prepared request settings.
         */
        static build(instance: Httpr, method: HttpMethod, url: string, params?: PlainObject, headers?: StringMap, body?: any): HttpRequestSettings;

        /**
         * Receive success response and apply interceptors.
         *
         * @returns {Promise<HttpResponse>} resolution handler.
         */
        static onSuccess(instance: Httpr, response: HttpResponse): Promise<HttpResponse>;

        /**
         * Receive error response and apply interceptors.
         *
         * @returns {Promise<HttpResponse>} rejection handler.
         */
        static onError(instance: Httpr, response: HttpResponse): Promise<HttpResponse>;
    }

    export class Httpr {

        /**
         * List of interceptors.
         */
        protected _interceptors: Array<HttprInterceptor>;

        /**
         * Httpr configuration.
         */
        protected _config: HttprConfig;

        /**
         * Httpr constructor.
         *
         * @param {HttprConfig} config - custom http configuration.
         */
        constructor(config: HttprConfig);

        /**
         * Return interceptors from this http instance.
         *
         * @returns {Array<HttprInterceptor>} list of interceptors.
         */
        readonly interceptors: Array<HttprInterceptor>;

        /**
         * Return a copy of current configuration.
         *
         * @returns {HttprConfig} current config.
         */
        readonly config: HttprConfig;

        /**
         * Perform a GET request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        get(url: string, params?: PlainObject, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform a POST request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {*} body - request body data.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        post(url: string, params?: PlainObject, body?: any, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform a PUT request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {*} body - request body data.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        put(url: string, params?: PlainObject, body?: any, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform a DELETE request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        del(url: string, params?: PlainObject, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform an OPTIONS request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        options(url: string, params?: PlainObject, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform a PATCH request.
         *
         * @see {@link Httpr.request}
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @returns {Promise<HttpResponse>} Promise handler.
         */
        patch(url: string, params?: PlainObject, headers?: StringMap): Promise<HttpResponse>;

        /**
         * Perform an HTTP Request.
         *
         * @see {@link HttprStatic}
         * @see {@link HttprProvider}
         *
         * @param {HttpMethod} method - http method to use.
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {StringMap} headers - hash of request headers to set.
         * @param {*} body - request body data.
         * @returns {Promise<HttpResponse>} promise handler.
         */
        request(method: HttpMethod, url: string, params?: PlainObject, headers?: StringMap, body?: any): Promise<HttpResponse>;

        /**
         * Add an interceptor to the Http instance.
         *
         * @param {HttprInterceptor} interceptor - instance of the interceptor to add.
         * @returns {Httpr} current Http instance.
         */
        register(interceptor: HttprInterceptor): Httpr;

        /**
         * Remove an interceptor from the Http instance.
         *
         * @param {?HttprInterceptor} interceptor - instance of the interceptor to remove.
         * If none is specified, remove all interceptors.
         * @returns {Httpr} current Http instance.
         */
        unregister(interceptor?: HttprInterceptor): Httpr;
    }

    /**
     * Form body for URL Encoded data media.
     */

    export class Form extends HttprBody<PlainObject> {
        constructor(data: PlainObject);

        /**
         * Serialize body content as string.
         *
         * @return {string} body.
         */
        toString(): string;
    }

    /**
     * Json body for application json data media.
     */

    export class Json extends HttprBody<any> {
        constructor(data: any);

        /**
         * Serialize body content as string.
         *
         * @return {string} body.
         */
        toString(): string;
    }
}

declare module 'httpr' { export = Httpr; }